@@ -26,12 +26,6 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
-#include <libintl.h>
-#define _(s) gettext(s)
-#include <netinet/in.h>
-#include <linux/socket.h>
-#include <linux/limits.h>
-#include <arpa/inet.h>
 #include <linux/capability.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -39,6 +33,7 @@
 #include <unistd.h>
 #include <sys/apparmor.h>
 
+#include "lib.h"
 #include "parser.h"
 #include "profile.h"
 #include "parser_yacc.h"
@@ -48,7 +43,7 @@
 /* #define DEBUG */
 #ifdef DEBUG
 #undef PDEBUG
-#define PDEBUG(fmt, args...) printf("Lexer: " fmt, ## args)
+#define PDEBUG(fmt, args...) fprintf(stderr, "Lexer: " fmt, ## args)
 #else
 #undef PDEBUG
 #define PDEBUG(fmt, args...)	/* Do nothing */
@@ -112,6 +107,7 @@ struct keyword_table {
 static struct keyword_table keyword_table[] = {
 	/* network */
 	{"network",		TOK_NETWORK},
+	{"unix",		TOK_UNIX},
 	/* misc keywords */
 	{"capability",		TOK_CAPABILITY},
 	{"if",			TOK_IF},
@@ -142,6 +138,7 @@ static struct keyword_table keyword_table[] = {
 	{"pivot_root",		TOK_PIVOTROOT},
 	{"in",			TOK_IN},
 	{"dbus",		TOK_DBUS},
+	{"signal",		TOK_SIGNAL},
 	{"send",                TOK_SEND},
 	{"receive",             TOK_RECEIVE},
 	{"bind",                TOK_BIND},
@@ -149,6 +146,9 @@ static struct keyword_table keyword_table[] = {
 	{"write",               TOK_WRITE},
 	{"eavesdrop",		TOK_EAVESDROP},
 	{"peer",		TOK_PEER},
+	{"trace",		TOK_TRACE},
+	{"tracedby",		TOK_TRACEDBY},
+	{"readby",		TOK_READBY},
 
 	/* terminate */
 	{NULL, 0}
@@ -183,7 +183,7 @@ static struct keyword_table rlimit_table[] = {
 };
 
 /* for alpha matches, check for keywords */
-static int get_table_token(const char *name __unused, struct keyword_table *table,
+static int get_table_token(const char *name unused, struct keyword_table *table,
 			   const char *keyword)
 {
 	int i;
@@ -226,255 +226,58 @@ int get_rlimit(const char *name)
 	return get_table_token("rlimit", rlimit_table, name);
 }
 
-struct network_tuple {
-	const char *family_name;
-	unsigned int family;
-	const char *type_name;
-	unsigned int type;
-	const char *protocol_name;
-	unsigned int protocol;
-};
-
-/* FIXME: currently just treating as a bit mask this will have to change
- * set up a table of mappings, there can be several mappings for a
- * given match.
- * currently the mapping does not set the protocol for stream/dgram to
- * anything other than 0.
- *   network inet tcp -> network inet stream 0 instead of
- *   network inet raw tcp.
- * some entries are just provided for completeness at this time
- */
-/* values stolen from /etc/protocols - needs to change */
-#define RAW_TCP 6
-#define RAW_UDP 17
-#define RAW_ICMP 1
-#define RAW_ICMPv6 58
-
-/* used by af_name.h to auto generate table entries for "name", AF_NAME
- * pair */
-#define AA_GEN_NET_ENT(name, AF) \
-	{name, AF, "stream",    SOCK_STREAM,    "", 0xffffff}, \
-	{name, AF, "dgram",     SOCK_DGRAM,     "", 0xffffff}, \
-	{name, AF, "seqpacket", SOCK_SEQPACKET, "", 0xffffff}, \
-	{name, AF, "rdm",       SOCK_RDM,       "", 0xffffff}, \
-	{name, AF, "raw",       SOCK_RAW,       "", 0xffffff}, \
-	{name, AF, "packet",    SOCK_PACKET,    "", 0xffffff},
-/*FIXME: missing {name, AF, "dccp", SOCK_DCCP, "", 0xfffffff}, */
-
-static struct network_tuple network_mappings[] = {
-	/* basic types */
-	#include "af_names.h"
-/* FIXME: af_names.h is missing AF_LLC, AF_TIPC */
-	/* mapped types */
-	{"inet",	AF_INET,	"raw",		SOCK_RAW,
-	 "tcp",	        1 << RAW_TCP},
-	{"inet",	AF_INET,	"raw",		SOCK_RAW,
-	 "udp",		1 << RAW_UDP},
-	{"inet",	AF_INET,	"raw",		SOCK_RAW,
-	 "icmp",	1 << RAW_ICMP},
-	{"inet",	AF_INET,	"tcp",		SOCK_STREAM,
-	 "",		0xffffffff},	/* should we give raw tcp too? */
-	{"inet",	AF_INET,	"udp",		SOCK_DGRAM,
-	 "",		0xffffffff},	/* should these be open masks? */
-	{"inet",	AF_INET,	"icmp",		SOCK_RAW,
-	 "",		1 << RAW_ICMP},
-	{"inet6",	AF_INET6,	"tcp",		SOCK_STREAM,
-	 "",		0xffffffff},
-	{"inet6",	AF_INET6,	"udp",		SOCK_DGRAM,
-	 "",		0xffffffff},
-/* what do we do with icmp on inet6?
-	{"inet6",	AF_INET,	"icmp",		SOCK_RAW,	0},
-	{"inet6",	AF_INET,	"icmpv6",	SOCK_RAW,	0},
-*/
-	/* terminate */
-	{NULL, 0, NULL, 0, NULL, 0}
-};
-
-/* The apparmor kernel patches up until 2.6.38 didn't handle networking
- * tables with sizes > AF_MAX correctly.  This could happen when the
- * parser was built against newer kernel headers and then used to load
- * policy on an older kernel.  This could happen during upgrades or
- * in multi-kernel boot systems.
- *
- * Try to detect the running kernel version and use that to determine
- * AF_MAX
- */
-#define PROC_VERSION "/proc/sys/kernel/osrelease"
-static size_t kernel_af_max(void) {
-	char buffer[32];
-	int major;
-	int fd, res;
-
-	if (!net_af_max_override) {
-		return 0;
-	}
-	/* the override parameter is specifying the max value */
-	if (net_af_max_override > 0)
-		return net_af_max_override;
-
-	fd = open(PROC_VERSION, O_RDONLY);
-	if (!fd)
-		/* fall back to default provided during build */
-		return 0;
-	res = read(fd, &buffer, sizeof(buffer) - 1);
-	close(fd);
-	if (res <= 0)
-		return 0;
-	buffer[res] = '\0';
-	res = sscanf(buffer, "2.6.%d", &major);
-	if (res != 1)
-		return 0;
-
-	switch(major) {
-	case 24:
-	case 25:
-	case 26:
-		return 34;
-	case 27:
-		return 35;
-	case 28:
-	case 29:
-	case 30:
-		return 36;
-	case 31:
-	case 32:
-	case 33:
-	case 34:
-	case 35:
-		return 37;
-	case 36:
-	case 37:
-		return 38;
-	/* kernels .38 and later should handle this correctly so no
-	 * static mapping needed
-	 */
-	default:
-		return 0;
-	}
-}
-
-/* Yuck. We grab AF_* values to define above from linux/socket.h because
- * they are more accurate than sys/socket.h for what the kernel actually
- * supports. However, we can't just include linux/socket.h directly,
- * because the AF_* definitions are protected with an ifdef KERNEL
- * wrapper, but we don't want to define that because that can cause
- * other redefinitions from glibc. However, because the kernel may have
- * more definitions than glibc, we need make sure AF_MAX reflects this,
- * hence the wrapping function.
- */
-size_t get_af_max() {
-	size_t af_max;
-	/* HACK: declare that version without "create" had a static AF_MAX */
-	if (!perms_create && !net_af_max_override)
-		net_af_max_override = -1;
-
-#if AA_AF_MAX > AF_MAX
-	af_max = AA_AF_MAX;
-#else
-	af_max = AF_MAX;
-#endif
-
-	/* HACK: some kernels didn't handle network tables from parsers
-	 * compiled against newer kernel headers as they are larger than
-	 * the running kernel expected.  If net_override is defined check
-	 * to see if there is a static max specified for that kernel
-	 */
-	if (net_af_max_override) {
-		size_t max = kernel_af_max();
-		if (max && max < af_max)
-			return max;
-	}
-
-	return af_max;
-}
-struct aa_network_entry *new_network_ent(unsigned int family,
-					 unsigned int type,
-					 unsigned int protocol)
-{
-	struct aa_network_entry *new_entry;
-	new_entry = (struct aa_network_entry *) calloc(1, sizeof(struct aa_network_entry));
-	if (new_entry) {
-		new_entry->family = family;
-		new_entry->type = type;
-		new_entry->protocol = protocol;
-		new_entry->next = NULL;
-	}
-	return new_entry;
-}
-
-struct aa_network_entry *network_entry(const char *family, const char *type,
-				       const char *protocol)
-{
-	int i;
-	struct aa_network_entry *new_entry, *entry = NULL;
-
-	for (i = 0; network_mappings[i].family_name; i++) {
-		if (family) {
-			PDEBUG("Checking family %s\n", network_mappings[i].family_name);
-			if (strcmp(family, network_mappings[i].family_name) != 0)
-				continue;
-			PDEBUG("Found family %s\n", family);
-		}
-		if (type) {
-			PDEBUG("Checking type %s\n", network_mappings[i].type_name);
-			if (strcmp(type, network_mappings[i].type_name) != 0)
-				continue;
-			PDEBUG("Found type %s\n", type);
-		}
-		if (protocol) {
-			PDEBUG("Checking protocol type %s\n", network_mappings[i].protocol_name);
-			if (strcmp(type, network_mappings[i].protocol_name) != 0)
-				continue;
-			/* fixme should we allow specifying protocol by #
-			 * without needing the protocol mapping? */
-		}
-		/* if here we have a match */
-		new_entry = new_network_ent(network_mappings[i].family,
-					    network_mappings[i].type,
-					    network_mappings[i].protocol);
-		if (!new_entry)
-			yyerror(_("Memory allocation error."));
-		new_entry->next = entry;
-		entry = new_entry;
-	}
-
-	return entry;
-};
-
 char *processunquoted(const char *string, int len)
 {
-	char *tmp, *s;
-	int l;
+	char *buffer, *s;
 
-	tmp = (char *)malloc(len + 1);
-	if (!tmp)
+	s = buffer = (char *) malloc(len + 1);
+	if (!buffer)
 		return NULL;
 
-	s = tmp;
-	for (l = 0; l < len; l++) {
-		if (string[l] == '\\' && l < len - 3) {
-			if (strchr("0123", string[l + 1]) &&
-			    strchr("0123456789", string[l + 2]) &&
-			    strchr("0123456789", string[l + 3])) {
-				/* three digit octal */
-				int res = (string[l + 1] - '0') * 64 +
-				    	  (string[l + 2] - '0') * 8 +
-					  (string[l + 3] - '0');
-				*s = res;
-				l += 3;
-			} else {
-				*s = string[l];
-			}
-			s++;
+	while (len > 0) {
+		const char *pos = string + 1;
+		long c;
+		if (*string == '\\' && len > 1 &&
+		    (c = strn_escseq(&pos, "", len)) != -1) {
+			/* catch \\ or \134 and other aare special chars and
+			 * pass it through to be handled by the backend
+			 * pcre conversion
+			 */
+			if (strchr("*?[]{}^,\\", c) != NULL) {
+				*s++ = '\\';
+				*s++ = c;
+			} else
+				*s++ = c;
+			len -= pos - string;
+			string = pos;
 		} else {
-			*s = string[l];
-			s++;
+			/* either unescaped char OR
+			 * unsupported escape sequence resulting in char being
+			 * copied.
+			 */
+			*s++ = *string++;
+			len--;
 		}
 	}
-
 	*s = 0;
 
-	return tmp;
+	return buffer;
+}
+
+/* rewrite a quoted string substituting escaped characters for the
+ * real thing.  Strip the quotes around the string */
+char *processquoted(const char *string, int len)
+{
+	/* skip leading " and eat trailing " */
+	if (*string == '"') {
+		len -= 2;
+		if (len < 0)	/* start and end point to same quote */
+			len = 0;
+		return processunquoted(string + 1, len);
+	}
+
+	/* no quotes? treat as unquoted */
+	return processunquoted(string, len);
 }
 
 char *processid(const char *string, int len)
@@ -487,72 +290,6 @@ char *processid(const char *string, int len)
 	return processunquoted(string, len);
 }
 
-/* rewrite a quoted string substituting escaped characters for the
- * real thing.  Strip the quotes around the string */
-
-char *processquoted(const char *string, int len)
-{
-	char *tmp, *s;
-	int l;
-	/* the result string will be shorter or equal in length */
-	tmp = (char *)malloc(len + 1);
-	if (!tmp)
-		return NULL;
-
-	s = tmp;
-	for (l = 1; l < len - 1; l++) {
-		if (string[l] == '\\' && l < len - 2) {
-			switch (string[l + 1]) {
-			case 't':
-				*s = '\t';
-				l++;
-				break;
-			case 'n':
-				*s = '\n';
-				l++;
-				break;
-			case 'r':
-				*s = '\r';
-				l++;
-				break;
-			case '"':
-				*s = '"';
-				l++;
-				break;
-			case '\\':
-				*s = '\\';
-				l++;
-				break;
-			case '0': case '1': case '2': case '3':
-				if ((l < len - 4) &&
-				    strchr("0123456789", string[l + 2]) &&
-				    strchr("0123456789", string[l + 3])) {
-					/* three digit octal */
-					int res = (string[l + 1] - '0') * 64 +
-					    (string[l + 2] - '0') * 8 +
-					    (string[l + 3] - '0');
-					*s = res;
-					l += 3;
-					break;
-				}
-				/* fall through */
-			default:
-				/* any unsupported escape sequence results in all
-				   chars being copied. */
-				*s = string[l];
-			}
-			s++;
-		} else {
-			*s = string[l];
-			s++;
-		}
-	}
-
-	*s = 0;
-
-	return tmp;
-}
-
 /* strip off surrounding delimiters around variables */
 char *process_var(const char *var)
 {
@@ -596,7 +333,7 @@ int str_to_boolean(const char *value)
 
 static int warned_uppercase = 0;
 
-static void warn_uppercase(void)
+void warn_uppercase(void)
 {
 	if (!warned_uppercase) {
 		pwarn(_("Uppercase qualifiers \"RWLIMX\" are deprecated, please convert to lowercase\n"
@@ -605,7 +342,7 @@ static void warn_uppercase(void)
 	}
 }
 
-static int parse_sub_mode(const char *str_mode, const char *mode_desc __unused)
+static int parse_sub_mode(const char *str_mode, const char *mode_desc unused)
 {
 
 #define IS_DIFF_QUAL(mode, q) (((mode) & AA_MAY_EXEC) && (((mode) & AA_EXEC_TYPE) != ((q) & AA_EXEC_TYPE)))
@@ -792,12 +529,12 @@ int parse_mode(const char *str_mode)
 	return mode;
 }
 
-static int parse_dbus_sub_mode(const char *str_mode, int *result, int fail, const char *mode_desc __unused)
+static int parse_X_sub_mode(const char *X, const char *str_mode, int *result, int fail, const char *mode_desc unused)
 {
 	int mode = 0;
 	const char *p;
 
-	PDEBUG("Parsing DBus mode: %s\n", str_mode);
+	PDEBUG("Parsing %s mode: %s\n", X, str_mode);
 
 	if (!str_mode)
 		return 0;
@@ -810,12 +547,12 @@ static int parse_dbus_sub_mode(const char *str_mode, int *result, int fail, cons
 reeval:
 		switch (current) {
 		case COD_READ_CHAR:
-			PDEBUG("Parsing DBus mode: found %s READ\n", mode_desc);
+			PDEBUG("Parsing %s mode: found %s READ\n", X, mode_desc);
 			mode |= AA_DBUS_RECEIVE;
 			break;
 
 		case COD_WRITE_CHAR:
-			PDEBUG("Parsing DBus mode: found %s WRITE\n",
+			PDEBUG("Parsing %s mode: found %s WRITE\n", X,
 			       mode_desc);
 			mode |= AA_DBUS_SEND;
 			break;
@@ -827,16 +564,16 @@ reeval:
 			switch (lower) {
 			case COD_READ_CHAR:
 			case COD_WRITE_CHAR:
-				PDEBUG("Parsing DBus mode: found invalid upper case char %c\n",
-				       current);
+				PDEBUG("Parsing %s mode: found invalid upper case char %c\n",
+				       X, current);
 				warn_uppercase();
 				current = lower;
 				goto reeval;
 				break;
 			default:
 				if (fail)
-					yyerror(_("Internal: unexpected DBus mode character '%c' in input"),
-						current);
+					yyerror(_("Internal: unexpected %s mode character '%c' in input"),
+						X, current);
 				else
 					return 0;
 				break;
@@ -846,21 +583,21 @@ reeval:
 		p++;
 	}
 
-	PDEBUG("Parsed DBus mode: %s 0x%x\n", str_mode, mode);
+	PDEBUG("Parsed %s mode: %s 0x%x\n", X, str_mode, mode);
 
 	*result = mode;
 	return 1;
 }
 
-int parse_dbus_mode(const char *str_mode, int *mode, int fail)
+int parse_X_mode(const char *X, int valid, const char *str_mode, int *mode, int fail)
 {
 	*mode = 0;
-	if (!parse_dbus_sub_mode(str_mode, mode, fail, ""))
+	if (!parse_X_sub_mode(X, str_mode, mode, fail, ""))
 		return 0;
-	if (*mode & ~AA_VALID_DBUS_PERMS) {
+	if (*mode & ~valid) {
 		if (fail)
-			yyerror(_("Internal error generated invalid DBus perm 0x%x\n"),
-				  mode);
+			yyerror(_("Internal error generated invalid %s perm 0x%x\n"),
+				X, mode);
 		else
 			return 0;
 	}
@@ -938,30 +675,6 @@ void free_cod_entries(struct cod_entry *list)
 	free(list);
 }
 
-void free_mnt_entries(struct mnt_entry *list)
-{
-	if (!list)
-		return;
-	if (list->next)
-		free_mnt_entries(list->next);
-	free(list->mnt_point);
-	free(list->device);
-	free_value_list(list->dev_type);
-	free_value_list(list->opts);
-
-	free(list);
-}
-
-void free_dbus_entries(struct dbus_entry *list)
-{
-	if (!list)
-		return;
-	if (list->next)
-		free_dbus_entries(list->next);
-
-	free_dbus_entry(list);
-}
-
 static void debug_base_perm_mask(int mask)
 {
 	if (HAS_MAY_READ(mask))
@@ -1073,95 +786,6 @@ void __debug_capabilities(uint64_t capset, const char *name)
 	printf("\n");
 }
 
-/* Bleah C++ doesn't have non-trivial designated initializers so we just
- * have to make sure these are in order.  This means we are more brittle
- * but there isn't much we can do.
- */
-const char *sock_types[] = {
-	"none",		/* 0 */
-	"stream",	/* 1 [SOCK_STREAM] */
-	"dgram",	/* 2 [SOCK_DGRAM] */
-	"raw",		/* 3 [SOCK_RAW] */
-	"rdm",		/* 4 [SOCK_RDM] */
-	"seqpacket",	/* 5 [SOCK_SEQPACKET] */
-	"dccp",		/* 6 [SOCK_DCCP] */
-	"invalid",	/* 7 */
-	"invalid",	/* 8 */
-	"invalid",	/* 9 */
-	"packet",	/* 10 [SOCK_PACKET] */
-	/*
-	 * See comment above
-	*/
-};
-#define ALL_TYPES 0x43e
-
-/* another case of C++ not supporting non-trivial designated initializers */
-#undef AA_GEN_NET_ENT
-#define AA_GEN_NET_ENT(name, AF) name, /* [AF] = name, */
-
-static const char *network_families[] = {
-#include "af_names.h"
-};
-
-void __debug_network(unsigned int *array, const char *name)
-{
-	unsigned int count = sizeof(sock_types)/sizeof(sock_types[0]);
-	unsigned int mask = ~((1 << count) -1);
-	unsigned int i, j;
-	int none = 1;
-	size_t af_max = get_af_max();
-
-	for (i = AF_UNSPEC; i < af_max; i++)
-		if (array[i]) {
-			none = 0;
-			break;
-		}
-
-	if (none)
-		return;
-
-	printf("%s: ", name);
-
-	/* This can only be set by an unqualified network rule */
-	if (array[AF_UNSPEC]) {
-		printf("<all>\n");
-		return;
-	}
-
-	for (i = 0; i < af_max; i++) {
-		if (array[i]) {
-			const char *fam = network_families[i];
-			if (fam)
-				printf("%s ", fam);
-			else
-				printf("#%u ", i);
-
-			/* All types/protocols */
-			if (array[i] == 0xffffffff || array[i] == ALL_TYPES)
-				continue;
-
-			printf("{ ");
-
-			for (j = 0; j < count; j++) {
-				const char *type;
-				if (array[i] & (1 << j)) {
-					type = sock_types[j];
-					if (type)
-						printf("%s ", type);
-					else
-						printf("#%u ", j);
-				}
-			}
-			if (array[i] & mask)
-				printf("#%x ", array[i] & mask);
-
-			printf("} ");
-		}
-	}
-	printf("\n");
-}
-
-
 struct value_list *new_value_list(char *value)
 {
 	struct value_list *val = (struct value_list *) calloc(1, sizeof(struct value_list));
@@ -1228,6 +852,17 @@ void print_value_list(struct value_list *list)
 	}
 }
 
+void move_conditional_value(const char *rulename, char **dst_ptr,
+			    struct cond_entry *cond_ent)
+{
+	if (*dst_ptr)
+		yyerror("%s conditional \"%s\" can only be specified once\n",
+			rulename, cond_ent->name);
+
+	*dst_ptr = cond_ent->vals->value;
+	cond_ent->vals->value = NULL;
+}
+
 struct cond_entry *new_cond_entry(char *name, int eq, struct value_list *list)
 {
 	struct cond_entry *ent = (struct cond_entry *) calloc(1, sizeof(struct cond_entry));
@@ -1294,27 +929,82 @@ int test_str_to_boolean(void)
 int test_processunquoted(void)
 {
 	int rc = 0;
-	const char *teststring, *processedstring;
+	const char *teststring;
+	const char *resultstring;
 
 	teststring = "";
 	MY_TEST(strcmp(teststring, processunquoted(teststring, strlen(teststring))) == 0,
 			"processunquoted on empty string");
 
+	teststring = "\\1";
+	resultstring = "\001";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on one digit octal");
+
+	teststring = "\\8";
+	resultstring = "\\8";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on invalid octal digit \\8");
+
+	teststring = "\\18";
+	resultstring = "\0018";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on one digit octal followed by invalid octal digit");
+
+	teststring = "\\1a";
+	resultstring = "\001a";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on one digit octal followed by hex digit a");
+
+	teststring = "\\1z";
+	resultstring = "\001z";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on one digit octal follow by char z");
+
+	teststring = "\\11";
+	resultstring = "\011";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on two digit octal");
+
+	teststring = "\\118";
+	resultstring = "\0118";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on two digit octal followed by invalid octal digit");
+
+	teststring = "\\11a";
+	resultstring = "\011a";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on two digit octal followed by hex digit a");
+
+	teststring = "\\11z";
+	resultstring = "\011z";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on two digit octal followed by char z");
+
+	teststring = "\\111";
+	resultstring = "\111";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on three digit octal");
+
+	teststring = "\\378";
+	resultstring = "\0378";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on three digit octal two large, taken as 2 digit octal plus trailing char");
+
 	teststring = "123\\421123";
-	MY_TEST(strcmp(teststring, processunquoted(teststring, strlen(teststring))) == 0,
-			"processunquoted on invalid octal");
+	resultstring = "123\0421123";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on two character octal followed by valid octal digit \\421");
 
-/* Oh wow, our octal processing is busticated - FIXME
 	teststring = "123\\109123";
-	processedstring = "123\109123";
-	MY_TEST(strcmp(processedstring, processunquoted(teststring, strlen(teststring))) == 0,
-			"processunquoted on octal 10");
+	resultstring = "123\109123";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on octal 109");
 
-	teststring = "123\\189123";
-	processedstring = "123\189123";
-	MY_TEST(strcmp(processedstring, processunquoted(teststring, strlen(teststring))) == 0,
-			"processunquoted on octal 10");
-*/
+	teststring = "123\\1089123";
+	resultstring = "123\1089123";
+	MY_TEST(strcmp(resultstring, processunquoted(teststring, strlen(teststring))) == 0,
+			"processunquoted on octal 108");
 
 	return rc;
 }
@@ -1353,7 +1043,7 @@ int test_processquoted(void)
 	free(out);
 
 	teststring = "\"a\\\\bcdefg\"";
-	processedstring = "a\\bcdefg";
+	processedstring = "a\\\\bcdefg";
 	out = processquoted(teststring, strlen(teststring));
 	MY_TEST(strcmp(processedstring, out) == 0,
 			"processquoted on quoted slash");
@@ -1401,19 +1091,30 @@ int test_processquoted(void)
 			"processquoted on quoted octal \\176");
 	free(out);
 
-	/* yes, our octal processing is lame; patches accepted */
-	teststring = "\"abc\\42defg\"";
-	processedstring = "abc\\42defg";
+	teststring = "\"abc\\429defg\"";
+	processedstring = "abc\0429defg";
 	out = processquoted(teststring, strlen(teststring));
 	MY_TEST(strcmp(processedstring, out) == 0,
-			"processquoted passthrough quoted invalid octal \\42");
+			"processquoted passthrough quoted invalid octal \\429");
 	free(out);
 
-	teststring = "\"abcdefg\\04\"";
-	processedstring = "abcdefg\\04";
+	teststring = "\"abcdefg\\4\"";
+	processedstring = "abcdefg\004";
 	out = processquoted(teststring, strlen(teststring));
 	MY_TEST(strcmp(processedstring, out) == 0,
-			"processquoted passthrough quoted invalid trailing octal \\04");
+			"processquoted passthrough quoted one digit trailing octal \\4");
+
+	teststring = "\"abcdefg\\04\"";
+	processedstring = "abcdefg\004";
+	out = processquoted(teststring, strlen(teststring));
+	MY_TEST(strcmp(processedstring, out) == 0,
+			"processquoted passthrough quoted two digit trailing octal \\04");
+
+	teststring = "\"abcdefg\\004\"";
+	processedstring = "abcdefg\004";
+	out = processquoted(teststring, strlen(teststring));
+	MY_TEST(strcmp(processedstring, out) == 0,
+			"processquoted passthrough quoted three digit trailing octal \\004");
 	free(out);
 
 	return rc;
