@@ -26,20 +26,18 @@
 #include <stdarg.h>
 #include <search.h>
 #include <string.h>
-#include <libintl.h>
 #include <errno.h>
-#define _(s) gettext(s)
 
 #include "parser.h"
 #include "profile.h"
-#include "mount.h"
-#include "dbus.h"
 #include "parser_yacc.h"
 
 /* #define DEBUG */
 #ifdef DEBUG
-#define PDEBUG(fmt, args...) printf("Lexer: " fmt, ## args)
+#undef PDEBUG
+#define PDEBUG(fmt, args...) fprintf(stderr, "Lexer: " fmt, ## args)
 #else
+#undef PDEBUG
 #define PDEBUG(fmt, args...)	/* Do nothing */
 #endif
 #define NPDEBUG(fmt, args...)	/* Do nothing */
@@ -70,7 +68,7 @@ void add_hat_to_policy(Profile *prof, Profile *hat)
 	}
 }
 
-static int add_entry_to_x_table(Profile *prof, char *name)
+int add_entry_to_x_table(Profile *prof, char *name)
 {
 	int i;
 	for (i = (AA_EXEC_LOCAL >> 10) + 1; i < AA_EXEC_COUNT; i++) {
@@ -192,29 +190,10 @@ void post_process_file_entries(Profile *prof)
 	}
 }
 
-void post_process_mnt_entries(Profile *prof)
+void post_process_rule_entries(Profile *prof)
 {
-	struct mnt_entry *entry;
-
-	list_for_each(prof->mnt_ents, entry) {
-		if (entry->trans) {
-			unsigned int mode = 0;
-			int n = add_entry_to_x_table(prof, entry->trans);
-			if (!n) {
-				PERROR("Profile %s has too many specified profile transitions.\n", prof->name);
-				exit(1);
-			}
-
-			if (entry->allow & AA_USER_EXEC)
-				mode |= SHIFT_MODE(n << 10, AA_USER_SHIFT);
-			if (entry->allow & AA_OTHER_EXEC)
-				mode |= SHIFT_MODE(n << 10, AA_OTHER_SHIFT);
-			entry->allow = ((entry->allow & ~AA_ALL_EXEC_MODIFIERS) |
-				       (mode & AA_ALL_EXEC_MODIFIERS));
-
-			entry->trans = NULL;
-		}
-	}
+	for (RuleList::iterator i = prof->rule_ents.begin(); i != prof->rule_ents.end(); i++)
+		(*i)->post_process(*prof);
 }
 
 
@@ -264,14 +243,10 @@ int load_policy(int option)
 	return load_policy_list(policy_list, option);
 }
 
-int load_hats(sd_serialize *p, Profile *prof)
+int load_hats(std::ostringstream &buf, Profile *prof)
 {
 	for (ProfileList::iterator i = prof->hat_table.begin(); i != prof->hat_table.end(); i++) {
-		if (!sd_serialize_profile(p, *i, 0)) {
-			PERROR(_("ERROR in profile %s, failed to load\n"),
-			       (*i)->name);
-			return -EINVAL;
-		}
+		sd_serialize_profile(buf, *i, 0);
 	}
 
 	return 0;
